-- stub module to add your own rules.
module UserRules (userRules) where

import List (nub,intersperse)
import StandardRules(Rule,Tag) -- gives some examples 
import RuleUtils -- useful to have a look at this too

{- datatype that rules manipulate :-


data Data = D {	name :: Name,			 -- type's name
			constraints :: [(Class,Var)], 
			vars :: [Var],		 -- Parameters
			body :: [Body],
			derives :: [Class],	 -- derived classes
			statement :: Statement}  -- type of statement
	   | Directive				 --|
	   | TypeName Name			 --| used by derive (ignore)
		deriving (Eq,Show) 

data Body = Body { constructor :: Constructor,
		    labels :: [Name], -- [] for a non-record datatype.
		    types :: [Type]} deriving (Eq,Show) 

data Statement = DataStmt | NewTypeStmt deriving (Eq,Show)

type Name = String
type Var = String
type Class = String
type Constructor = String

type Rule = (Tag, Data->Doc)

-}

-- add your rules to this list
userRules :: [Rule]
userRules = [
    ("ATermConvertible", atermfn),
    ("Binary", binary),
    ("Typeable", typeablefn),
    ("Term", dyntermfn),
    ("OldTerm", termfn),
    ("Implode", implodefn),
    ("Explode", explodefn),
    ("Observable", observablefn),
    ("HFoldable", hfoldfn),
    ("Gfu", gfufn),
    ("Gfoldl", gfoldlfn),
    ("Gmap", gmapfn)
    ]




-- useful helper things
namesupply   = [text [x,y] | x <- ['a' .. 'z'], 
                             y <- ['a' .. 'z'] ++ ['A' .. 'Z']]
mknss []     _  = []
mknss (c:cs) ns =
  let (thisns,rest) = splitAt (length (types c)) ns
  in thisns: mknss cs rest 

mkpattern :: Constructor -> [a] -> [Doc] -> Doc
mkpattern c l ns =
  if null l then text c
  else parens (hsep (text c : take (length l) ns))

instanceheader cls dat = instanceheader' cls dat []
-- things get a little bit more complicated with Term instances:
-- to enable unambiguous resolution of overlapping in the presence
-- of Term instances for String and [a], we need to add extra
-- constraints to the instance header if the datatype in question
-- uses its parameter types in lists; dyntermfn calls this helper
-- function directly, supplying the type parameters used in such way
instanceheader' cls dat lv =
  let fv     = vars dat
      tycon  = name dat
      ctx    = map (\v-> text cls <+> text v)
      lctx   = map (\v-> text cls <+> text ("["++v++"]"))
      parenSpace = parens . hcat . sepWith space
  in
  hsep [ text "instance"
       , opt fv (\v -> parenList (ctx v++lctx lv) <+> text "=>")
       , text cls
       , opt1 (texts (tycon: fv)) parenSpace id
       , text "where"
       ]

doublequote str
  = "\""++str++"\""

mkList :: [Doc] -> Doc
mkList xs = text "[" <> hcat (punctuate comma xs) <> text "]"


-- TODO Constructors!
{- old, does not work for parameterized types
typeablefn :: Data -> Doc
typeablefn  dat  = tcname <+> equals <+> text "mkTyCon" <+> text (doublequote $ name dat) $$
    instanceheader "Typeable" dat $$ block (
	[text "typeOf _ = mkAppTy"  <+> tcname <+> text "[]"]) where
    tcname = text ("_tc_" ++ (name dat)  ++ "Tc")
-}

typeablefn :: Data -> Doc
typeablefn  dat  
  = tcname <+> equals <+> text "mkTyCon" <+> text (doublequote $ name dat) $$
    instanceheader "Typeable" dat $$ block (
	[ text "typeOf x = mkAppTy"  <+> 
	  tcname <+> 
	  text "[" <+> hcat (sepWith comma (map getV' (vars dat))) <+> text "]" $$ 
	  wheres ]) 
    where
      tcname = text ("_tc_" ++ (name dat)  ++ "Tc")
      wheres = where_decls (map getV (vars dat))
      tpe    = text (name dat) <+> hcat (sepWith space (map text (vars dat)))
      getV' var
        = text "typeOf" <+> parens (text "get" <> text var <+> text "x")
      getV var 
        = text "get" <> text var <+> text "::" <+> tpe <+> text "->" <+> text var $$
          text "get" <> text var <+> equals <+> text "undefined"

where_decls [] = empty 
where_decls ds = text "  where" $$ block ds

dyntermfn :: Data -> Doc
dyntermfn dat = instanceheader' "Term" dat lvs $$ block [ 
    text "explode (x::"<>a<>text ") = TermRep (toDyn x, f x, g x) where", block (
	zipWith f cvs cs ++ zipWith g cvs cs
	)] where
	    f cv c = text "f" <+> ppCons cv c <+> equals <+> mkList (map (text "explode" <+>) $ vrs c cv)
	    g cv c = text "g" <+> ppCons underscores c <+> text "xs" <+> 
--		text "|" <+> mkList (vrs c cv) <+> text "<- TermRep.fArgs xs" <+> equals <+> text "toDyn" <+> parens (parens (text (constructor c) <+> hsep (map h (vrs c cv))) <> text "::a" ) 
		equals <+> text "case TermRep.fArgs xs of" <+> mkList (vrs c cv) <+> text "->" <+> text "toDyn" <+> parens (parens (text (constructor c) <+> hsep (map h (vrs c cv))) <> text "::"<>a<>text "" ) 
	    h n = parens $ text "TermRep.fDyn" <+> n
	    cvs = mknss cs namesupply
	    cs = body dat
	    vrs c cv = take (length (types c)) cv
	    underscores = repeat $ text "_"
	    a = text (name dat) <+> hcat (sepWith space (map text (vars dat)))
	    lvs = nub $ concatMap (lists (vars dat)) $ concatMap types $ body dat
	    lists ps (Arrow a b)    = lists ps a ++ lists ps b    
	    -- a type application might recursively require list constraints
	    -- we can't decide that locally, so we err on the safe side..
	    lists ps (Apply a b)    = [ s | Var s <- [b], s `elem` ps ] 
                                ++ lists ps a ++ lists ps b
	    lists ps (LApply a bs)  = [ s | Var s <- bs, s `elem` ps ] 
                                ++ lists ps a ++ concatMap (lists ps) bs
	    lists ps (Tuple as)     = concatMap (lists ps) as
	    lists ps (List (Var s)) = [ s | s `elem` ps ]
	    lists ps (List a)       = lists ps a
	    lists ps _                           = []

-- begin observable 

observablefn :: Data -> Doc
observablefn  dat = 
  let cs  = body dat
      cvs = mknss cs namesupply
  in
  instanceheader "Observable" dat $$ 
  block (zipWith observefn cvs cs)

observefn cv c = 
    text "observer" <+> ppCons cv c <+> text "= send"  <+> text (doublequote (constructor c)) <+> parens (text "return" <+> text (constructor c) <+> hsep (map f (take (length (types c)) cv))) where
    f n = text "<<" <+> n

-- begin here for Binary derivation


binary dat = 
  let cs  = body dat
      cvs = mknss cs namesupply
      k   = (ceiling . logBase 2 . realToFrac . length) cs
  in
  instanceheader "Binary" dat $$
  block (  zipWith3 (putfn k) [0..] cvs cs
        ++ getfn k [0..] cvs cs
        :  getFfn k [0..] cvs cs
        :  zipWith (sizefn k) cvs cs
        )

putfn k n cv c =
  text "put bh" <+> ppCons cv c <+> text "= do" $$
  nest 8 (
    text "pos <- putBits bh" <+> text (show k) <+> text (show n) $$
    vcat (map (text "put bh" <+>) cv) $$
    text "return pos"
  )

ppCons cv c = mkpattern (constructor c) (types c) cv

getfn k ns cvs cs =
  text "get bh = do" $$
  nest 8 (
    text "h <- getBits bh" <+> text (show k) $$
    text "case h of" $$
    nest 2 ( vcat $
      zipWith3 (\n vs c-> text (show n) <+> text "-> do" $$
                          nest 6 (
                            vcat (map (\v-> v <+> text "<-" <+> text "get bh") vs) $$
                            text "return" <+> ppCons vs c
                          ))
               ns cvs cs
    )
  )

getFfn k ns cvs cs =
  text "getF bh p =" <+>
  nest 8 (
    text "let (h,p1) = getBitsF bh 1 p in" $$
    text "case h of" $$
    nest 2 ( vcat $
      zipWith3 (\n vs c-> text (show n) <+> text "->" <+>
                          parens (cons c <> text ",p1") <+>
                          hsep (map (\_-> text "<< getF bh") vs))
               ns cvs cs
    )
  )
  where cons =  text . constructor

sizefn k [] c =
  text "sizeOf" <+> ppCons [] c <+> text "=" <+> text (show k)
sizefn k cv c =
  text "sizeOf" <+> ppCons cv c <+> text "=" <+> text (show k) <+> text "+" <+>
  hsep (intersperse (text "+") (map (text "sizeOf" <+>) cv))


-- end of binary derivation



-- begin of implode derivation
-- Author: Joost.Visser@cwi.nl

termfn dat
  = instanceSkeleton "Term" 
       [ (makeImplode (name dat),defaultImplode (name dat))
       , (makeGetTypeId (name dat),defaultGetTypeId (name dat))
       , (makeExplode (name dat),defaultExplode)
       ] 
       dat

implodefn dat
  = instanceSkeleton "Implode" 
       [ (makeImplode (name dat),defaultImplode (name dat))
       , (makeGetTypeId (name dat),defaultGetTypeId (name dat))
       ] 
       dat


makeImplode name body
  = let cvs = head (mknss [body] namesupply)
    in text "implode" <+> 
       text "(TermRep" <+> text "(TypeRep" <+> text (doublequote name) <+> text "[])" <+>
--       text "(TermRep" <+> text (doublequote name) <+>
       text (doublequote (constructor body)) <+>
       text "[" <+> 
       hcat (intersperse (text ",") cvs) <+> 
       text "])" <+>
       text "=" <+> text "let" <+>
       vcat (map childImplode cvs) <+>
       text "in" <+>
       ppCons (map addPrime cvs) body
defaultImplode name
  = hsep $ texts ["implode", "u", "=", "implodeError", (doublequote name), "u"]
childImplode v
  = (addPrime v) <+> text "=" <+> text "implode" <+> v

addPrime doc = doc <> (text "'")

{- Old monadic variant:
makeImplode name body
  = let cvs = head (mknss [body] namesupply)
    in text "implode" <+> 
       text "(TermRep" <+> text (doublequote name) <+>
       text (doublequote (constructor body)) <+>
       text "[" <+> 
       hcat (intersperse (text ",") cvs) <+> 
       text "])" <+>
       text "=" <+> text "do" <+>
       vcat ( (map childImplode cvs)++[text "return"] ) <+>
       ppCons cvs body
defaultImplode
  = hsep $ texts ["implode", "_", "=" ,"mzero"]
childImplode v
  = v <+> text "<-" <+> text "implode" <+> v
-}

makeGetTypeId name body
  = empty
defaultGetTypeId name
--  = text "getTypeId" <+>
  = text "getTypeRep" <+>
    text "_" <+>
    text "=" <+>
    text "TypeRep" <+> text (doublequote name) <+> text "[]"


-- end of implode derivation




-- begin of explode derivation 
-- Author: Joost.Visser@cwi.nl

explodefn dat
  = instanceSkeleton "Explode" [(makeExplode (name dat),defaultExplode)] dat

makeExplode name body
  = let cvs = head (mknss [body] namesupply)
    in text "explode" <+> 
       ppCons cvs body <+>
       text "=" <+>
--       text "(TermRep" <+> text (doublequote name) <+>
       text "(TermRep" <+> text "(TypeRep" <+> text (doublequote name) <+> text "[])" <+>
       text (doublequote (constructor body)) <+>
       text "[" <+> 
       hcat (intersperse (text ",") (map childExplode cvs)) <+> 
       text "])"
defaultExplode
  = empty

childExplode v
  = text "explode" <+> v

-- end of explode derivation


-- begin of ATermConvertible derivation 
-- Author: Joost.Visser@cwi.nl

atermfn dat
  = instanceSkeleton "ATermConvertible" 
      [ (makeToATerm (name dat),defaultToATerm)
      , (makeFromATerm (name dat),defaultFromATerm (name dat))
      ] 
      dat

makeToATerm name body
  = let cvs = head (mknss [body] namesupply)
    in text "toATerm" <+> 
       ppCons cvs body <+>
       text "=" <+>
       text "(AAppl" <+>
       text (doublequote (constructor body)) <+>
       text "[" <+> 
       hcat (intersperse (text ",") (map childToATerm cvs)) <+> 
       text "])"
defaultToATerm
  = empty
childToATerm v
  = text "toATerm" <+> v

makeFromATerm name body
  = let cvs = head (mknss [body] namesupply)
    in text "fromATerm" <+> 
       text "(AAppl" <+>
       text (doublequote (constructor body)) <+>
       text "[" <+> 
       hcat (intersperse (text ",") cvs) <+> 
       text "])" <+>
       text "=" <+> text "let" <+>
       vcat (map childFromATerm cvs) <+>
       text "in" <+>
       ppCons (map addPrime cvs) body
defaultFromATerm name
  = hsep $ texts ["fromATerm", "u", "=", "fromATermError", (doublequote name), "u"]
childFromATerm v
  = (addPrime v) <+> text "=" <+> text "fromATerm" <+> v

-- end of ATermConvertible derivation



-- begin of HFoldable derivation 
-- Author: Joost Visser and Ralf Laemmel

hfoldfn dat
  = instanceSkeleton "HFoldable" 
      [ (make_hfoldr (name dat), default_hfoldr),
        (make_conof (name dat), default_conof)
      ] 
      dat

make_hfoldr name body
  = let cvs = head (mknss [body] namesupply)
    in text "hfoldr'" <+> 
       text "alg" <+>
       ppCons cvs body <+>
       text "=" <+>
       foldl (\rest var -> text "hcons alg" <+> var  <+> parens rest) 
             (text "hnil alg" <+> text (constructor body))
             cvs 

default_hfoldr
  = empty

-- end of HFoldable derivation


-- begin of Gfu derivation 
-- Author: Ralf Laemmel
-- Status: Experimental

gfufn dat
  = myinstanceSkeleton "Term" 
      [ (make_gfoldl (name dat),default_gfoldl),
        (make_gunfold (name dat),default_gunfold),
        (make_conof (name dat), default_conof)
      ] 
      [make_consof]
      dat


make_consof bodies
  =  text "consOf" <+> text "_" <+> text "=" <+> 
     text ("[" ++ perbodies bodies ++ "]")
 where
  perbodies [] = []
  perbodies (b:bs) = one b ++ foldl cons [] bs
  one = doublequote . constructor
  cons r b' = r ++ ", " ++ one b'

-- end of Gfu derivation



-- begin of Gfoldl derivation 
-- Author: Ralf Laemmel
-- Status: Experimental

gfoldlfn dat
  = instanceSkeleton "Term" 
      [ (make_gfoldl (name dat),default_gfoldl),
        (make_conof (name dat), default_conof)
      ] 
      dat


make_conof name body
  = let cvs = head (mknss [body] namesupply)
    in text "conOf" <+> 
       ppCons cvs body <+>
       text "=" <+>
       text (doublequote (constructor body))

default_conof
  = empty


-- end of Gfoldl derivation


-- begin of Gmap derivation 
-- Author: Ralf Laemmel
-- Status: Experimental

gmapfn dat
  = instanceSkeleton "Term" 
      [ (make_gfoldl (name dat),default_gfoldl),
        (make_gmapT (name dat),default_gmapT),
        (make_conof (name dat), default_conof)
      ] 
      dat

make_gfoldl name body
  = let cvs = head (mknss [body] namesupply)
    in text "gfoldl" <+> 
       text "f" <+> text "z" <+>
       ppCons cvs body <+>
       text "=" <+>
       foldl (\rest var ->  rest <+> text "`f`" <+> var) 
             (text "z" <+> text (constructor body))
             cvs 
default_gfoldl
  = empty

make_gunfold name body
  = let cvs = head (mknss [body] namesupply)
    in text "gunfold" <+> 
       text "f" <+> text "z" <+> text "_" <+>
       text (doublequote (constructor body)) <+>
       text "=" <+>
       foldl (\rest _ ->  text "f" <+> parens rest) 
             (text "z" <+> text (constructor body))
             cvs

default_gunfold
  = text "gunfold" <+> 
    text "_" <+> text "_" <+> text "e" <+> text "_" <+>
    text "=" <+> text "e"



-- Adopted from RuleUtils.hs
-- Ralf Laemmel
type IFunction' = [Body] -> Doc -- instance function


-- Adopted from RuleUtils.hs
-- Ralf Laemmel
myinstanceSkeleton :: Class -> [(IFunction,Doc)] -> [IFunction'] -> Data -> Doc
myinstanceSkeleton s ii ii' d
  = (simpleInstance s d <+> text "where") $$ block functions
 where
  functions = (concatMap f ii) ++ (map f' ii')
  f (i,dflt) = map i (body d) ++ [dflt]
  f' i       = i (body d)

make_gmapT name body
  = let cvs = head (mknss [body] namesupply)
    in text "gmapT" <+> 
       text "f" <+>
       ppCons cvs body <+>
       text "=" <+>
       foldl (\rest var ->  rest <+> (parens (text "f" <+> var))) 
             (text (constructor body))
             cvs 
default_gmapT
  = empty

-- end of Gmap derivation
