{-+
Instance declarations in the source code are assigned names and
added to the instance database, which gets used during context reduction.
-}
module TiInstanceDB(IDB,Instance,emptyIdb,extendIdb,findInst,instType) where
import TiTypes(Type,Subst(..),Types(..),funT,HsIdentI(..),Typing(..),kupscheme)
import TiSolve()
import Unification(match)
import Maybe(mapMaybe,isJust)
--import HsIdent -- hmm
import PrettyPrint
import SpecialNames
import TiDefinedNames(definedTypeName,optDefinedTypeName)
import FiniteMap
import MUtils(( # ))
--import IOExts(trace) -- debug

{-+
The instance database is simpliy a list of instances. An instance like

    #instance (Show a,Show b) => Show (Either a b)#

might be represented in the database as

    #(Show (Either a b),(inst_Show_Either,[Show a,Show b]))#
-}

type Instance i = (Type i,(i,[Type i]))
instClass (hd,_) = definedTypeName hd

newtype IDB i = Idb (FiniteMap i [Instance i]) --deriving Show

instType ks (c,(v,ctx)) = HsVar v:>:kupscheme ks (funT (ctx++[c]))

fakeInstType (c,(_,ctx)) = (tv t,t) -- vars in the same order as in instType...
  where t = funT (ctx++[c])

emptyIdb = Idb emptyFM
--extendIdb1 inst (Idb idb) = Idb (inst:idb)
extendIdb insts (Idb idb) = Idb (addListToFM_C (++) idb cinsts)
  where
    cinsts = [(instClass i,[i])|i<-insts]

--namesIdb (Idb idb) = [dn|(_,(dn,_))<-idb]

findInst (Idb idb) inst =
     --trace (pp$"Searching "<+>length insts<+>"instances") $
     map instantiate $ mostSpecific $ mapMaybe matchInst insts
  where
    insts = maybe allInsts (lookupWithDefaultFM idb []) $
	      optDefinedTypeName inst
      where
        -- Used when looking for an instance in an unknown class!
        allInsts = concat (eltsFM idb)

    matchInst i@(ip,_) = (,) i # match [(ip,inst)]

    instantiate (i@(ip,(dn,ips)),s) = 
      ((dn,apply (S s) ips),(fakeInstType i,S s))

-- Support for overlapping instances:
mostSpecific is = pick [] is
  where
    pick is1 [] = is1
    pick is1 (i:is2) =
      if all (i `moreSpecificThan`) is1 && all (i `moreSpecificThan`) is2
      then [i]
      else pick (i:is1) is2

    i1 `moreSpecificThan` i2 = isJust (match [(instHead i2,instHead i1)])
      where instHead ((ih,_),_) = ih

----
instance Show i => Show (IDB i) where
  showsPrec _ (Idb insts) = shows (fmToList insts)

instance (IsSpecialName i,Printable i) => Printable (IDB i) where
  ppi (Idb insts) = vcat [pinst i|is<-eltsFM insts,i<-is]

pinst (t,(dn,ctx)) = dn<+>"= instance"<+>ctx<+>"=>"<+>t
