module PfeAlfaCmds(pfeAlfaCmds) where
import Prelude hiding (putStrLn,writeFile,catch)
import Monad(filterM)
import Maybe(mapMaybe)
import List(intersect)

import HsModule

import RemovePatBinds(remPats)
import RemoveListCompProp(rmAllListComp)
import SimpFunBind(simpAllFunBind)
import SimpPatMatch(simpAllPatMatch)
import DefinedNames(addName)
import PFE4(rewriteAndTypeCheck)
import PFE0(moduleList,subGraph,newerThan)
import PfeParse(moduleArgs)
import Pfe0Cmds(getSortedSubGraph)
import TI((+++))
import TiModule(joinModules,representative)

import Prop2Alfa(transModule,modPath)
import BaseStruct2Alfa(packageSynonym,joinModuleNames)
--import FileConv(printModule)
import qualified UAbstract as U

import AbstractIO
import DirUtils(getModificationTimeMaybe)
import MUtils
import EnvM(withEnv)

pfeAlfaCmds =
  [--("alfa1",(moduleArgs prop2alfa1,"translate modules to Alfa")),
   ("alfa",(moduleArgs prop2alfa1,"translate modules to Alfa")),
   ("alfa2",(moduleArgs prop2alfa2,"translate modules to Alfa"))]

prop2alfa1 = prop2alfa' ("alfa1",rewrite1)
prop2alfa2 = prop2alfa' ("alfa2",rewrite2)

prop2alfa' rewrite@(rwname,_) ms =
    do dir <- getEnv "APFE_DIR" `catch` const (return "alfa")
       changedsccs <- changedSccs dir =<< getSortedSubGraph ms
       let ms = concat changedsccs
       if null ms then done
         else do tms <- rewriteAndTypeCheck rewrite (Just ms)
                 mapM_ (transSCC dir tms) changedsccs
  where
    transSCC dir alltms ms = mapM (writeModule dir . trans1) ms
      where
        joinedmn = joinModuleNames ms
	rm = representative ms
	tms = [(m,tm)|(m,(_,(tms,_)))<-alltms,m `elem` ms,(rwn,tm)<-tms,rwn==rwname]
	joinedtm = joinModules (map snd tms)

        trans1 m = (m,U.Module transm)
          where
            transm = if m==rm
	             then transds ++ syn
	             else imp:syn

            U.Module transds = withEnv (ms,env) (transModule joinedtm)
            syn = if m==joinedmn then [] else packageSynonym m joinedmn
	    imp = U.ImportDecl (U.Import (modPath rm))

        env = foldr ((+++).snd.snd.snd.snd) ([],[]) alltms

rewrite1 = simpAllPatMatch . simpAllFunBind . rewrite2
rewrite2 = remPats . rmAllListComp . addName

{-
prop2alfa1 =
    mapM_ writeModule.transModules @@ rewriteAndTypeCheck rewrite.Just
  where
    transModules ms = mapMaybe transModule' ms
      where
        transModule' (n,(_,(optm,_))) =
	    do m <- optm
	       return (n,withEnv ([n],env) (transModule m))
        env = foldr ((+++).snd.snd.snd.snd) ([],[]) ms
-}

writeModule dir (n,U.Module decls) =
    do let path=dir++"/"++modPath n
       ePutStrLn ("Updating: "++path)
       writeFile path ({-printModule-}show m)
  where
    m = U.Module (prefix++decls)
    prefix =
      [ U.Comment magic,
	U.ImportDecl (U.Import "Haskell.alfa"),
        U.Comment "{-# Alfa hidetypeannots on #-}"]

    magic = "-- Automatically converted from Haskell by hs2alfa..."

changedSccs dir g =
    -- A quick hack to avoid retranslating unchanged modules...
    do changed <- map (fst.snd) # filterM moduleChanged (concat g)
       let allsccs = [[m|(f,(m,is))<-scc]|scc<-g]
	   sccs= [scc|scc<-allsccs,
		      not.null $
		      moduleList (subGraph g scc) `intersect` changed]
       return sccs
  where
    moduleChanged (path,(m,_)) =
      newerThan  # getModificationTime path
                <# getModificationTimeMaybe (dir++"/"++modPath m)

