module Pfe4Cmds where

import PfeParse(moduleArgs,just,qualIds)
import Pfe3Cmds(pfe3Cmds)

import PFE0(pput,epput)
import PFE2(getExports)
import PFE4(runPFE4Cmds,topTypes,typeCheck,rewriteAndTypeCheck,modEnv)
import RemovePatBinds(remPats)
import RemoveListComp(rmAllListComp)
import SimpPatMatch(simpAllPatMatch)
import SimpFunBind(simpAllFunBind)
import Ents(Ent(..))
import QualNames(mkUnqual,unqual)
import Relations(relToList)
import TypedIds(isValue)
import TI(Typing(..),envFrom,topVal,topType)
import HasBaseName(getBaseName)
import HsIdent(accHsIdent2,getHSName)
import SrcLoc(srcLoc)

import MUtils(( # ))
import PrettyPrint(pp,vcat,(<+>),(<>),($$),nest,ppi,sep,kw,con,tcon,modn)
import PrettySymbols(el)
import HsDeclPretty(ppContext)
import Products((><))
import OpTypes(cmpBy)
import List(partition,sortBy)

{-+
The type checker can return different types, so to avoid an ambiguity when
the result is used only for pretty printing, the function tcOutput can be passed
in to restrict the result type.
-}

pfe4 ext tcOutput = runPFE4Cmds ext (pfe4Cmds tcOutput)

pfe4Cmds tcOutput=
    pfe3Cmds ++
    [("tc",       (tcmd tc,       "type check and display decorated modules")),
     ("tcpb",     (tcmd tcpb,     "remove pattern bindings, then tc")),
     ("tclc",     (tcmd tclc,     "remove list comprehensions, then tc")),
     ("tcpm",     (tcmd tcpm,     "simplify patterns, then tc")),
     ("types",    (tcmd types,    "show types/kinds of top-level entities")),
     ("typeof",   (tqcmd typeof,  "show types of named top-level entities")),
     ("kindof",   (tqcmd kindof,  "show kinds of named top-level entities")),
     ("instances",(tcmd instances,"list instances defined in a module")),
     ("iface",    (tcmd ppIfaces, "show the interfaces of modules")),
     ("usedtypes",(tcmd utypes,   "show what types identifers are used at"))]
  where
    tcmd f = moduleArgs (f tcOutput)
    tqcmd f = qualIds (f tcOutput)

instances tcOutput ms =
  do pfe4info <- tcOutput # topTypes (Just ms)
     pput.vcat $ [ppContext ps<+>p |(m,(_,(_,(insts,_))))<-pfe4info,
		                    m `elem` ms,
		                    (p,(_,ps))<-insts]

types tcOutput ms =
  do pfe4info <- tcOutput # topTypes (Just ms)
     pput.vcat $ [map (fmap fst) ks$$ts |(m,(_,(_,(_,(ks,ts)))))<-pfe4info,
		                    m `elem` ms]

typeof tcOutput qids =
  do pfe4info <- tcOutput # topTypes (Just (map fst qids))
     mapM_ (typeof1 pfe4info) qids
{-
typeof1 pfe4info (m,s) = pput (vcat bs)
  where
    q=topVal m s
    bs = [b|(m',(_,(_,(_,(ks,ts)))))<-pfe4info,
		m'==m,
		b@(x:>:_)<-ts,
		getHSName x==q]
-}
kindof tcOutput qids =
  do pfe4info <- tcOutput # topTypes (Just (map fst qids))
     mapM_ (kindof1 pfe4info) qids
{-
kindof1 pfe4info (m,s) = pput (vcat bs)
  where
    q=topType m s
    bs = [fmap fst b|(m',(_,(_,(_,(ks,ts)))))<-pfe4info,
		     m'==m,
		     b@(x:>:_)<-ks,
		     getHSName x==q]
-}
typeof1 = info1 snd id
kindof1 = info1 fst fst

info1 envsel infosel pfe4info (m,s) = pput (vcat bs)
  where
    q=topType m s
    bs = [fmap infosel b|Just env<-[modEnv pfe4info m],
		     b@(x:>:_)<-envsel env,
		     getHSName x==q]

utypes tcOutput ms =
  do pfe4info <- tcOutput # typeCheck (Just ms)
     mapM_ (utypes1' pfe4info) ms
  where
    utypes1' pfe4info mn =
        maybe (epput $ "Unkown module:"<+>mn) utypes1
		      (lookup "id".fst.snd=<<lookup mn pfe4info)
      where
        utypes1 m =
          pput (mn<>":"$$
		nest 2 (vcat.map u.sortBy (cmpBy fst).map pos.snd.envFrom $ m))

        pos xt@(x:>:_) = (srcLoc x,xt)

        u (pos,x:>:(sc,optt)) =
	    sep [x<+>"at"<+>pos<+>"::",
		 nest 4 (maybe (ppi sc) ppi optt)]

tc tcOutput ms = pptc ms =<< (tcOutput # typeCheck (just ms))

pptc ms pfe4info =
  pput.vcat $ [tm |(m,(_,(tms,_)))<-pfe4info, m `elem` ms,("id",tm)<-tms]

tcpb tcOutput ms =
  pptc ms =<< (tcOutput # rewriteAndTypeCheck ("tcpb",remPats) (just ms))

tclc tcOutput ms =
  pptc ms =<< (tcOutput # rewriteAndTypeCheck ("tclc",rmAllListComp) (just ms))

tcpm tcOutput ms =
  pptc ms =<< (tcOutput # rewriteAndTypeCheck ("tcpm",simpAllPatMatch.simpAllFunBind) (just ms))

ppIfaces tcOutput ms =
  do exports <- getExports (Just ms)
     pfe4info <- tcOutput # topTypes (Just ms)
     mapM_ (pput.ppIface.iface exports pfe4info) ms

moduleInterface m =
  do exports <- getExports (Just [m])
     types <- topTypes (Just [m])
     return (iface exports types m)

iface expRels pfe4info m =
    (m,((types><kinds).partition isValue.map snd.relToList.snd
        # lookup m expRels))
  where
    kinds = map (info fst fst)
    types = map (info snd id)

    info envsel infosel e@(Ent m' n _) = (e,i)
      where
        Just env = conv . envsel # modEnv pfe4info m'
	i = lift n (infosel # lookup (mkUnqual(getBaseName n)) env)


    conv env = [(unqual' (getBaseName x),y)|x:>:y<-env]
    unqual' = unqual `asTypeOf` id

    lift x = maybe (error $ pp $ "Not found:"<+>x) id
      
ppIface (m,Nothing) = "Unknown module:"<+>m
ppIface (m,Just (ts,ks)) =
    kw "module"<+>modn m<>":" $$
    nest 2 (vcat (map (ppInfo ppT) ks) $$ "" $$ vcat (map (ppInfo ppV) ts))
  where
    ppV = accHsIdent2 ppi con
    ppT = accHsIdent2 ppi tcon
    ppInfo ppId (Ent m' n _,info) = {-ppName m'-} ppId n<+>el<+>info
    {-
    ppName m' n =
      if m'==m
      then ppi n
      else m'<>"."<>n
    -}
